# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Ansible Project

import pytest
from packaging.version import Version as PypiVer
from semantic_version import Version as SemVer

from antsibull_core.dependency_files import BuildFile

SIMPLE_TEST_FILE = """_ansible_version: 4
_ansible_core_version: 2.11.0rc1
community.general: >=1.0.0,<2.0.0
community.routeros: >=2.0.0-a2,<3.0.0
"""

SIMPLE_TEST_DEPS = {
    "community.general": SemVer("1.0.0"),
    "community.routeros": SemVer("2.0.0-a2"),
}

SIMPLE_TEST_FILE_2 = """_ansible_version: 6
_ansible_core_version: 2.13.0rc1
community.general: >=1.0.0,<2.0.0
community.routeros: >=2.0.0-a2,<3.0.0
"""

SIMPLE_TEST_DEPS_2 = {
    "community.general": SemVer("1.0.0"),
    "community.routeros": SemVer("2.0.0-a2"),
}


@pytest.mark.parametrize(
    "ansible_ver, core_ver, dependencies, file_contents",
    (
        ("4.0.0", "2.11.0rc1", SIMPLE_TEST_DEPS, SIMPLE_TEST_FILE),
        ("6.0.0", "2.13.0rc1", SIMPLE_TEST_DEPS_2, SIMPLE_TEST_FILE_2),
    ),
)
def test_build_file_write(tmpdir, ansible_ver, core_ver, dependencies, file_contents):
    filename = tmpdir / "test.build"
    bf = BuildFile(filename)
    bf.write(PypiVer(ansible_ver), core_ver, dependencies)

    with open(filename) as f:
        assert f.read() == file_contents



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2021, Ansible Project

"""
YAML handling.
"""

from __future__ import annotations

# pylint: disable=unused-import
from antsibull_fileutils.yaml import load_yaml_bytes  # noqa: F401
from antsibull_fileutils.yaml import load_yaml_file  # noqa: F401
from antsibull_fileutils.yaml import store_yaml_file  # noqa: F401
from antsibull_fileutils.yaml import store_yaml_stream  # noqa: F401



# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Ansible Project

import os

import pytest
import twiggy.levels

import antsibull_core.logging as al


def test_library_settings(capsys):
    """Importing without calling initialize_app_logging leaves logging disabled."""
    assert al.log.min_level == twiggy.levels.DISABLED

    al.log.error("test")
    captured = capsys.readouterr()
    assert captured.err == ""
    assert captured.out == ""


def test_initialize_app_logging(capsys):
    """Calling initialize_app_logging outputs WARNING and above to stderr."""
    al.initialize_app_logging()

    al.log.error("test")
    al.log.warning("second test")
    al.log.debug("third test")
    captured = capsys.readouterr()
    assert captured.err == "ERROR:antsibull|test\nWARNING:antsibull|second test\n"
    assert captured.out == ""


def test_initialize_app_logging_debug(capsys):
    os.environ["ANTSIBULL_EARLY_DEBUG"] = "1"
    al.initialize_app_logging()

    al.log.error("test")
    al.log.warning("second test")
    al.log.debug("third test")
    captured = capsys.readouterr()
    assert (
        captured.err
        == "ERROR:antsibull|test\nWARNING:antsibull|second test\nDEBUG:antsibull|third test\n"
    )
    assert captured.out == ""



# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2023 Maxwell G <maxwell@gtmx.me>

import contextlib
import os
import tempfile
from pathlib import Path

import nox

DEFAULT_MODE = os.environ.get("OTHER_ANTSIBULL_MODE", "auto")
IN_CI = "GITHUB_ACTIONS" in os.environ
ALLOW_EDITABLE = os.environ.get("ALLOW_EDITABLE", str(not IN_CI)).lower() in (
    "1",
    "true",
)

# Always install latest pip version
os.environ["VIRTUALENV_DOWNLOAD"] = "1"
nox.options.sessions = "lint", "test", "coverage"


def install(session: nox.Session, *args, editable=False, **kwargs):
    # nox --no-venv
    if isinstance(session.virtualenv, nox.virtualenv.PassthroughEnv):
        session.warn(f"No venv. Skipping installation of {args}")
        return
    # Don't install in editable mode in CI or if it's explicitly disabled.
    # This ensures that the wheel contains all of the correct files.
    if editable and ALLOW_EDITABLE:
        args = ("-e", *args)
    session.install(*args, "-U", **kwargs)


def other_antsibull(
    mode: str | None = None,
) -> list[str | Path]:
    if mode is None:
        mode = DEFAULT_MODE
    to_install: list[str | Path] = []
    args = ("antsibull-fileutils",)
    for project in args:
        path = Path("../", project)
        path_exists = path.is_dir()
        if mode == "auto":
            if path_exists:
                mode = "local"
            else:
                mode = "git"
        if mode == "local":
            if not path_exists:
                raise ValueError(f"Cannot install {project}! {path} does not exist!")
            if ALLOW_EDITABLE:
                to_install.append("-e")
            to_install.append(path)
        elif mode == "git":
            to_install.append(
                f"{project} @ "
                f"https://github.com/ansible-community/{project}/archive/main.tar.gz"
            )
        elif mode == "pypi":
            to_install.append(project)
        else:
            raise ValueError(f"install_other_antsibull: invalid argument mode={mode!r}")
    return to_install


@nox.session(python=["3.9", "3.10", "3.11", "3.12", "3.13", "3.14"])
def test(session: nox.Session):
    install(
        session,
        ".[test, coverage]",
        *other_antsibull(),
        editable=True,
    )
    covfile = Path(session.create_tmp(), ".coverage")
    more_args = []
    if session.python not in {"3.9", "3.10", "3.11"}:
        more_args.append("--error-for-skips")
    session.run(
        "pytest",
        "--cov-branch",
        "--cov=antsibull_core",
        "--cov-report",
        "term-missing",
        *more_args,
        *session.posargs,
        env={"COVERAGE_FILE": f"{covfile}", **session.env},
    )


@nox.session
def coverage(session: nox.Session):
    install(session, ".[coverage]", *other_antsibull(), editable=True)
    combined = map(str, Path().glob(".nox/*/tmp/.coverage"))
    # Combine the results into a single .coverage file in the root
    session.run("coverage", "combine", "--keep", *combined)
    # Create a coverage.xml for codecov
    session.run("coverage", "xml")
    # Display the combined results to the user
    session.run("coverage", "report", "-m")


@nox.session
def lint(session: nox.Session):
    session.notify("formatters")
    session.notify("codeqa")
    session.notify("typing")


@nox.session
def formatters(session: nox.Session):
    install(session, ".[formatters]", *other_antsibull())
    posargs = list(session.posargs)
    if IN_CI:
        posargs.append("--check")
    session.run("isort", *posargs, "src", "tests", "noxfile.py")
    session.run("black", *posargs, "src", "tests", "noxfile.py")


@nox.session
def codeqa(session: nox.Session):
    install(session, ".[codeqa]", *other_antsibull(), editable=True)
    session.run("flake8", "src", *session.posargs)
    session.run("pylint", "--rcfile", ".pylintrc.automated", "src/antsibull_core")
    session.run("reuse", "lint")
    session.run("antsibull-changelog", "lint")


@nox.session
def typing(session: nox.Session):
    # pyre does not work when we don't install ourself in editable mode 🙄.
    others = other_antsibull()
    install(session, "-e", ".[typing]", *others)
    session.run("mypy", "src/antsibull_core")

    additional_libraries = []
    for path in others:
        if isinstance(path, Path):
            additional_libraries.extend(("--search-path", str(path / "src")))

    purelib = (
        session.run(
            "python",
            "-c",
            "import sysconfig; print(sysconfig.get_path('purelib'))",
            silent=True,
        )
        or ""
    ).strip()
    platlib = (
        session.run(
            "python",
            "-c",
            "import sysconfig; print(sysconfig.get_path('platlib'))",
            silent=True,
        )
        or ""
    ).strip()
    session.run(
        "pyre",
        "--source-directory",
        "src",
        "--search-path",
        purelib,
        "--search-path",
        platlib,
        "--search-path",
        "stubs/",
        *additional_libraries,
    )


def check_no_modifications(session: nox.Session) -> None:
    modified = session.run(
        "git",
        "status",
        "--porcelain=v1",
        "--untracked=normal",
        external=True,
        silent=True,
    )
    if modified:
        session.error(
            "There are modified or untracked files. "
            "Commit, restore, or remove them before running this"
        )


@contextlib.contextmanager
def isolated_src(session: nox.Session):
    """
    Create an isolated directory that only contains the latest git HEAD
    """
    with tempfile.TemporaryDirectory() as _tmpdir:
        tmp = Path(_tmpdir)
        session.run(
            "git",
            "archive",
            "HEAD",
            f"--output={tmp / 'HEAD.tar'}",
            "--prefix=build/",
            external=True,
        )
        with session.chdir(tmp):
            session.run("tar", "-xf", "HEAD.tar", external=True)
        with session.chdir(tmp / "build"):
            yield


@nox.session
def bump(session: nox.Session):
    check_no_modifications(session)
    if len(session.posargs) not in (1, 2):
        session.error(
            "Must specify 1-2 positional arguments: nox -e bump -- <version> "
            "[ <release_summary_message> ]."
            " If release_summary_message has not been specified, "
            "a file changelogs/fragments/<version>.yml must exist"
        )
    version = session.posargs[0]
    fragment_file = Path(f"changelogs/fragments/{version}.yml")
    if len(session.posargs) == 1:
        if not fragment_file.is_file():
            session.error(
                f"Either {fragment_file} must already exist, "
                "or two positional arguments must be provided."
            )
    # Needs newer antsibull-changelog for hatch version auto-detection support
    install(session, "antsibull-changelog[toml] >= 0.26.0", "hatch")
    session.run("hatch", "version", version)
    if len(session.posargs) > 1:
        fragment = session.run(
            "python",
            "-c",
            "import sys, yaml ; "
            f"yaml.dump(dict(release_summary={repr(session.posargs[1])}), sys.stdout)",
            silent=True,
        )
        with open(fragment_file, "w") as fp:
            fp.write(fragment)
        session.run(
            "git",
            "add",
            "src/antsibull_core/__init__.py",
            str(fragment_file),
            external=True,
        )
        session.run("git", "commit", "-m", f"Prepare {version}.", external=True)
    session.run("antsibull-changelog", "release")
    session.run(
        "git",
        "add",
        "CHANGELOG.rst",
        "CHANGELOG.md",
        "changelogs/changelog.yaml",
        "changelogs/fragments/",
        # src/antsibull_core/__init__.py is not committed in the last step
        # when the release_summary fragment is created manually
        "src/antsibull_core/__init__.py",
        external=True,
    )
    install(session, ".")  # Smoke test
    session.run("git", "commit", "-m", f"Release {version}.", external=True)
    session.run(
        "git",
        "tag",
        "-a",
        "-m",
        f"antsibull-core {version}",
        "--edit",
        version,
        external=True,
    )
    dist = Path.cwd() / "dist"
    with isolated_src(session):
        session.run("hatch", "build", "--clean", str(dist))


@nox.session
def publish(session: nox.Session):
    check_no_modifications(session)
    install(session, "hatch")
    session.run("hatch", "publish", *session.posargs)
    version = session.run("hatch", "version", silent=True).strip()
    session.run("hatch", "version", "post")
    session.run("git", "add", "src/antsibull_core/__init__.py", external=True)
    session.run("git", "commit", "-m", "Post-release version bump.", external=True)



# Author: Toshio Kuratomi <tkuratom@redhat.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2021, Toshio Kuratomi
"""Schemas for app and lib contexts."""

import typing as t
from functools import cached_property

import pydantic as p

from ..utils.collections import ContextDict
from .config import DEFAULT_LOGGING_CONFIG, LoggingModel
from .validators import convert_bool, convert_none, convert_path


class BaseModel(p.BaseModel):
    """
    Configuration for all Context object classes.

    :cvar model_config: Sets the following information

        :arg allow_mutation: ``False``.  Prevents setattr on the contexts.
        :arg extra: ``p.Extra.forbid``.  Prevents extra fields on the contexts.
        :arg validate_all: ``True``.  Validates default values as well as user supplied ones.
    """

    model_config = p.ConfigDict(frozen=True, extra="forbid", validate_default=True)


class AppContext(BaseModel):
    """
    Structure and defaults of the app_ctx.

    :ivar extra: a mapping of arg/config keys to values.  Anything in here is unchecked by a
        schema.  These are usually leftover command line arguments and config entries. If
        values stored in extras need default values, they need to be set outside of the context
        or the entries can be given an actual entry in the AppContext to take advantage of the
        schema's checking, normalization, and default setting.
    :ivar logging_cfg: Configuration of the application logging.
    """

    model_config = p.ConfigDict(frozen=True, extra="allow", validate_default=True)

    @cached_property
    def extra(self) -> ContextDict:
        d = (self.__pydantic_extra__ or {}).get("extra", {})
        return ContextDict.validate_and_convert(d)

    logging_cfg: LoggingModel = LoggingModel.model_validate(DEFAULT_LOGGING_CONFIG)


class LibContext(BaseModel):
    """
    Structure and defaults of the lib_ctx.

    :ivar chunksize: number of bytes to read or write at one time for network or file IO
    :ivar process_max: Maximum number of worker processes for parallel operations.  It may be None
        to mean, use all available CPU cores.
    :ivar thread_max: Maximum number of helper threads for parallel operations
    :ivar file_check_content: Maximum number of bytes of a file to read before writing it to
        compare contents. If contents are as expected, file is not overwritten. Set to 0 to
        disable.
    :ivar max_retries: Maximum number of times to retry an http request (in case of timeouts and
        other transient problems.
    :ivar ansible_core_repo_url: Url to the ansible-core git repo.
    :ivar galaxy_url: URL of the galaxy server to get collection info from
    :ivar logging_cfg: Configuration of the application logging
    :ivar pypi_url: URL of the pypi server to query for information
    :ivar collection_cache: If set, must be a path pointing to a directory where collection
        tarballs are cached so they do not need to be downloaded from Galaxy twice.
    :ivar trust_collection_cache: If set to ``True``, will assume that if the collection
        cache contains an artifact, it is the current one available on the Galaxy server.
        This avoids making a request to the Galaxy server to figure out the artifact's
        checksum and comparing it before trusting the cached artifact.
    :ivar ansible_core_cache: If set, must be a path pointing to a directory where ansible-core
        tarballs are cached so they do not need to be downloaded from PyPI twice.
    :ivar trust_ansible_core_cache: If set to ``True``, will assume that if the ansible-core
        cache contains an artifact, it is the current one available on PyPI. This avoids making a
        request to PyPI to figure out the artifact's checksum and comparing it before trusting
        the cached artifact.
    """

    chunksize: int = 4096

    max_retries: int = 10
    process_max: t.Optional[int] = None
    thread_max: int = 8
    file_check_content: int = 262144
    ansible_core_repo_url: p.HttpUrl = p.HttpUrl("https://github.com/ansible/ansible/")
    galaxy_url: p.HttpUrl = p.HttpUrl("https://galaxy.ansible.com/")
    pypi_url: p.HttpUrl = p.HttpUrl("https://pypi.org/")
    collection_cache: t.Optional[str] = None
    trust_collection_cache: bool = False
    ansible_core_cache: t.Optional[str] = None
    trust_ansible_core_cache: bool = False

    # pylint: disable-next=unused-private-member
    __convert_nones = p.field_validator("process_max", mode="before")(convert_none)
    # pylint: disable-next=unused-private-member
    __convert_paths = p.field_validator(
        "ansible_core_cache", "collection_cache", mode="before"
    )(convert_path)
    # pylint: disable-next=unused-private-member
    __convert_bools = p.field_validator(
        "trust_ansible_core_cache", "trust_collection_cache", mode="before"
    )(convert_bool)
